// ============================================================================
//  (C)OPYRIGHT 2003-2004 by Sebastian Reuss und Andreas Schibilla
//
// ----------------------------------------------------------------------------
//  Projekt         : BVATisch
//  Modul/Unit      : Region Growing
//  Ersterstellung  : 09. Mrz 2004
//  Letzte nderung :
//  Autor           : Sebastian Reuss    (ii5062)
//                    Andreas Schibilla  (ii4900)
//  Sprache         : Borland Delphi 6.0240 Personal-Edition
// ============================================================================
//
// ============================================================================

library gruppe6;

{$R *.res}

uses
  Windows, MMSystem, SysUtils, Classes, Dialogs, Graphics, Math, ICVCImg, CVIMAGELib_TLB;

type
  // Informationsstruktur, die vom Plugin ausgefllt werden kann. Jeweils
  // Zeiger auf Zeichenketten (nullterminierte Strings).
  TInfoString = record
                  pFilterGroup: PChar;		   {Filtergruppe}
                  pFilterName: PChar;			{Name des Filters}
                  pFilterAuthor: PChar;      {Entwickler des Filter-Plugins}
                  pFilterVersion: PChar;     {Version}
                end;

  // 3x3 Filtermatrix
  PTFilterKern = ^TFilterKern;
  TFilterKern  = packed array [0..2,0..2] of Single;

  // Weitere optionale Parameter fr die Plugins
  PTRGBVektor  = ^TRGBVektor;
  TRGBVektor   = packed array [0..2] of Single;

  TRGB         = record
                  r,g,b  : pVPAT;
                 end;

  TPixel       = record
                  wert     : Byte;
                  todel    : Boolean;
                  Bereich  : Longint;
                  si       : Byte;
                 end;

  TBild        = array of array of TPixel;

  TTicks       = record
                  summe,
                  gradient,
                  grau,
                  markieren,
                  wachsen,
                  verschmelzen,
                  schwellen   : Cardinal
                 end;

 TStartpunktPtr= ^TStartpunkt;
 TStartpunkt   = record
                  x,y         : Longint;
                  wert        : Byte;
                  si          : Byte;
                  prev, next  : TStartpunktPtr;
                 end;
 TGradient     = ( Sobel, PseudoLaplace, Roberts);

 TBereich      = record
                  FarbwertSumme  : Longint;
                  AnzPunkte      : Longint;
                  id             : Longint;
                  Wachsbar       : Boolean;
                  First, Last    : TStartpunktPtr;
                 end;

var
  g_strStatus  : String;               //kompletter String fr die Textausgabe des Plugins
  g_ticks      : TTicks;               //Rechenzeiten, die fr die einzelnen Bearbeitungsschritte bentigt wurden
  g_bild       : TBild;                //selbsdefinierte Bildstruktur
  g_breite     : Longint;              //Bildbreite
  g_hoehe      : Longint;              //Bildhhe
  g_first      : TStartpunktPtr;       //Anfang der Startpunktliste
  g_last       : TStartpunktPtr;       //Ende der Startpunktliste
  g_bereiche   : array of TBereich;    //Array aller Bereiche

const
  C_LEERER_BEREICH           = -1;
  C_DEFAULT : TFilterKern    = ( (1,1,2), (15,0,20), (8,1,20) );
  C_MIN     : TFilterKern    = ( (1,1,1), (1,0,1), (1,1,1) );
  C_MAX     : TFilterKern    = ( (50,50,10), (255,2,255), (255,2,255) );
  C_INTK                     = 3;

// ============================================================================
//  Name      				: KopiereGraubild
// ----------------------------------------------------------------------------
//  Eingangsparameter : pScr1								| Zeiger auf das Quellbild
//  Ausgangsparameter : -                          | -
// ----------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_breite_, g_hoehe
// ----------------------------------------------------------------------------
//  Funktion          : Wenn das Quellbild ein Graubild ist, in selbstdefinierte
//                      Bildstruktur kopieren (g_bild)
//  Bemerkungen       : -
// ============================================================================
procedure KopiereGraubild(pSrc1:Pointer);
var
 lSrcBase, x, y   : longint;
 pSrcVPA          : pVPAT;
begin
 GetImageVPA(pSrc1, 0, Pointer(lSrcBase), Pointer(pSrcVPA) );
 // setze Gre des Arrays des Buffers fr das Bild
 g_breite   := ImageWidth(pSrc1);
 g_hoehe    := ImageHeight(pSrc1);
 SetLength(g_bild, g_breite, g_hoehe);

 {$R-}
 for x := 0 to g_breite-1 do
   for y := 0 to g_hoehe-1 do
   begin
      g_bild[x,y].wert := PByte(lSrcBase + pSrcVPA^[x].xEntry + pSrcVPA^[y].yEntry)^;
      g_bild[x,y].Bereich := C_LEERER_BEREICH;
   end;
 {$R+}
end;


// ============================================================================
//  Name      			 : KopiereFarbbild
// ----------------------------------------------------------------------------
//  Eingangsparameter : pScr1								| Zeiger auf das Quellbild
//  Ausgangsparameter : -                          | -
// ----------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_breite_, g_hoehe
// ----------------------------------------------------------------------------
//  Funktion          : Wenn das Quellbild ein RGB-Bild ist, erst in Graustufen
//                      umwandeln, dann in selbstdefinierte Bildstruktur kopieren (g_bild)
//  Bemerkungen       : -
// ============================================================================
procedure KopiereFarbbild(pSrc1:Pointer);
var
 lSrcBase      : longint;
 pSrcVPA       : TRGB;
 lDstBase      : longint;
 pDstVPA       : pVPAT;
 x,y           : longint;
begin
 // Bitplanes vom Quellbild ermitteln
 GetImageVPA(pSrc1, 0, Pointer(lSrcBase), Pointer(pSrcVPA.r) );
 GetImageVPA(pSrc1, 1, Pointer(lSrcBase), Pointer(pSrcVPA.g) );
 GetImageVPA(pSrc1, 2, Pointer(lSrcBase), Pointer(pSrcVPA.b) );

 // setze Gre des Arrays des Buffers fr das Bild
 g_breite   := ImageWidth(pSrc1);
 g_hoehe    := ImageHeight(pSrc1);
 SetLength(g_bild, g_breite, g_hoehe);

 // Zeitmessung
 g_ticks.grau := timeGetTime;

 // Umwandeln in Graustufen
 {$R-}
 for x := 0 to g_breite-1 do
   for y := 0 to g_hoehe-1 do
   begin
      g_bild[x,y].todel := True;
      g_bild[x,y].wert  := Round((PByte(lSrcBase + pSrcVPA.r^[x].xEntry + pSrcVPA.r^[y].yEntry)^ +
                                  PByte(lSrcBase + pSrcVPA.g^[x].xEntry + pSrcVPA.b^[y].yEntry)^ +
                                  PByte(lSrcBase + pSrcVPA.b^[x].xEntry + pSrcVPA.g^[y].yEntry)^) / 3);
      g_bild[x,y].Bereich := C_LEERER_BEREICH;
   end;
 {$R+}
 g_ticks.grau := timeGetTime - g_ticks.grau;
 g_strStatus  := g_strStatus+ #13#10+'Zeit fr Graustufenumwandlung (in ms):'+#09+#09+#09+IntToStr(g_ticks.grau);
end;


// ============================================================================
//  Name      			 : KopiereZielbild
// ----------------------------------------------------------------------------
//  Eingangsparameter : -  								| -
//  Ausgangsparameter : dst                        | Struktur des Zielbildes
// ----------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild
// ----------------------------------------------------------------------------
//  Funktion          : kopiert das (fertige) Zielbild aus der selbstdefinierten Bildstruktur
//                      in die vom Plugin vorgegebene Struktur
//  Bemerkungen       : -
// ============================================================================
procedure KopiereZielbild(var dst:TCVImage);
var
 lDstBase      : longint;
 pDstVPA       : pVPAT;
 x,y           : longint;
 bmp           : TBitmap;
 SaveDialog    : TSaveDialog;
begin
 // Leeres Ausgabe-Image erzeugen (1 Bit-Plane) und Bitplane emritteln
 dst.CreateEmptyImage(1, g_breite, g_hoehe, false);
 GetImageVPA(IMG(dst.Image), 0, Pointer(lDstBase), Pointer(pDstVPA) );
 bmp        := TBitmap.Create;
 bmp.Width  := g_breite;
 bmp.Height := g_hoehe;
 {$R-}
 for x := 0 to g_breite-1 do
   for y := 0 to g_hoehe-1 do
   begin
       PByte(lDstBase + pDstVPA^[x].xEntry + pDstVPA^[y].yEntry)^ := g_bild[x,y].wert;
       bmp.Canvas.Pixels[x,y] := RGB(g_bild[x,y].wert,g_bild[x,y].wert,g_bild[x,y].wert);
   end;
 {$R+}
 SaveDialog := TSaveDialog.Create(nil);
 SaveDialog.DefaultExt := 'bmp';
 SaveDialog.Title := 'Sie knnen nun das Zielbild als RGB-Bitmap speichern';
 SaveDialog.Filter := 'Bitmaps (*.bmp)|*.bmp|All files (*.*)|*.*';
 SaveDialog.FilterIndex := 1;
 if SaveDialog.Execute then
   bmp.SaveToFile(SaveDialog.FileName);
 bmp.Free;
 SaveDialog.Free;
end;


// ============================================================================
//  Name      			 : BerechneIntSchwellen
// ----------------------------------------------------------------------------
//  Eingangsparameter : teilerB, teilerH				| Anzahl kacheln in Lnge und Breite
//                      k                          | vom benutzer bergebener Parameter zur
//                                                   Einstellung von si
//  Ausgangsparameter : -                          | -
// ----------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_breite_, g_hoehe
// ----------------------------------------------------------------------------
//  Funktion          : berechnet fr jede Kachel die Intensittsschwelle si
//                      Diese wird in jedem Pixel gespeichert
//  Bemerkungen       : -
// ============================================================================
procedure BerechneIntSchwellen(teilerB, teilerH : Longint; k : Byte);
var
 kx,ky, x,y    : Longint;
 min,max       : byte;
 si            : Byte;
begin
 // Zeitmessung
 g_ticks.schwellen := timeGetTime;
 if teilerB > g_breite then
   teilerH := g_breite;
 if teilerH > g_hoehe then
   teilerH := g_hoehe;

 for ky:=0 to teilerH-1 do     // durchlaufe Kacheln in y Richtung
   for kx:=0 to teilerB-1 do       // durchlaufe Kacheln in x Richtung
   begin
         min := 255;
         max := 0;

         // Berechne Intensittsmaxima und Minima
         for y:=Round(ky*g_hoehe/teilerH) to Round((ky+1)*g_hoehe/teilerH)-1 do     // durchlaufe Pixel der Kachel in y Richtung
            for x:=Round(kx*g_breite/teilerB) to Round((kx+1)*g_breite/teilerB)-1 do     // durchlaufe Pixel der Kachel in x Richtung
            begin
               if g_bild[x,y].wert > max then
                  max := g_bild[x,y].wert;
               if g_bild[x,y].wert < min then
                  min := g_bild[x,y].wert;
            end;
         si := Round( (max-min) / k );
         // Jeder Pixel bekommt si-Wert
         for y:=Round(ky*g_hoehe/teilerH) to Round((ky+1)*g_hoehe/teilerH)-1 do     // durchlaufe Pixel der Kachel in y Richtung
            for x:=Round(kx*g_breite/teilerB) to Round((kx+1)*g_breite/teilerB)-1 do     // durchlaufe Pixel der Kachel in x Richtung
               g_bild[x,y].si := si;
   end;
 g_ticks.schwellen := timeGetTime - g_ticks.schwellen;
 g_strStatus  := g_strStatus+ #13#10+'Zeit fr Berechnung der Intensittsschwellen (in ms):'+#09+IntToStr(g_ticks.schwellen);
end;


// =================================================================================================
//  Funktion          : benutzt einen Gradientoperator fr die Berechnung von Randpunkten, die dann
//                      zum Lschen markiert werden (kommen nicht als Startpunkte in Frage)
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : delatS          - Toleranz, Abweichung DeltaS (in Prozent)
//                      operator        - Welcher Operator soll fr das Kantenfinden
//                                        benutzt werden: (Sobel, PseudoLaplace, Roberts)
//  Ausgangsparameter : -
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_ticks.gradient, g_hoehe, g_breite, g_bild, g_strStatus
// =================================================================================================
procedure BerechneGradient(deltaS : Byte; operator : TGradient);
var
 x,y                 : Longint;
 op                  : Integer;
 a,b,c,d,e,f,g,h,i   : Byte;
begin
 // Zeitmessung
 g_ticks.gradient := timeGetTime;

 for y:=1 to g_hoehe-2 do                       // durchlaufe alle Bildpunkte (ohne Randpunkte)
   for x:=1 to g_breite-2 do
   begin
      a := g_bild[x-1,y-1].wert;                // jeder Punkt um den Ausgangspunkt bekommt einen Namen
      b := g_bild[x,y-1].wert;
      c := g_bild[x+1,y-1].wert;
      d := g_bild[x-1,y].wert;
      e := g_bild[x,y].wert;
      f := g_bild[x+1,y].wert;
      g := g_bild[x-1,y+1].wert;
      h := g_bild[x,y+1].wert;
      i := g_bild[x+1,y+1].wert;
      case operator of                          // berechne Operator mit entspr. Formel
         PseudoLaplace:    op := Abs(b+d+h+f-4*e);
         Sobel:            op := Abs( (a + 2*b + c) - (g + 2*h + i)) + Abs( (a + 2*d + g) - (c + 2*f + i));
         Roberts:          op := Max( Abs (a-e), Abs(b-d) );
      end;
      g_bild[x,y].todel := op > (deltaS/100 * g_bild[x,y].si);      // markiere Randpunkte zum Lschen
   end;
 g_ticks.gradient := timeGetTime - g_ticks.gradient;
 g_strStatus  := g_strStatus+ #13#10+'Zeit fr Berechnung der Gradienten (in ms):'+#09+#09+IntToStr(g_ticks.gradient);
end;


// =================================================================================================
//  Funktion          : Erstellt eine Liste mit allen mglichen Startpunkten (Initialpunkten)
//                      Ungltige Punkte (Randpunkte) mit Flag 'todel' werden nicht verwendet
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : -
//  Ausgangsparameter : -
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_first, g_last, g_bild, g_strStatus
// =================================================================================================
procedure ErzeugeListe;
var
 x,y : Longint;
 p   : TStartpunktPtr;
begin
 g_first := nil;
 g_last  := nil;
 for y:=g_hoehe-2 downto 1  do            // durchlaufe das Bild (ohne Randpunkte)
   for x:=g_breite-2 downto 1 do
      if not g_bild[x,y].todel then       // wenn Bildpunkt kein Randpunkt
      begin
         New(p);                          // erstelle neues Element und fge es am Anfang der Liste ein
         if (g_first <> nil) then
            g_first^.prev := p;
         p^.x     := x;                   // setze Werte des Elements in der Liste
         p^.y     := y;
         p^.wert  := g_bild[x,y].wert;
         p^.si    := g_bild[x,y].si;
         p^.prev  := nil;                 // biege die Listenzeiger um
         p^.next  := g_first;
         g_first := p;
         if g_last = nil then
            g_last := p;
      end;
end;


// =================================================================================================
//  Funktion          : Entfernt das Element 'el' aus der Startpunktliste und gibt den Speicher frei
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : el              - Element, das aus der Liste entfernt werden soll
//  Ausgangsparameter : first           - Zeiger auf den Anfang der Liste
//                      last            - zeiger auf das Ende der Liste
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : -
// =================================================================================================
procedure LoescheElement(el : TStartpunktPtr; var first : TStartpunktPtr; var last : TStartpunktPtr);
begin
 if el <> nil then                        // wenn neues Eleent nicht ungltig (nil)
 begin
   if el^.prev <> nil then                // gibt es einen Vorgnger?
      el^.prev^.next := el^.next
   else
   begin
      el^.next^.prev := nil;
      first := first^.next;
   end;

   if el^.next <> nil then                // gibt es einen Nachfolger?
      el^.next^.prev := el^.prev
   else
   begin
      el^.prev^.next := nil;
      last := last^.prev;
   end;

   Dispose(el);
 end;
end;


// =================================================================================================
//  Funktion          : Fgt ein Element am Ende der Startpunkte-Liste ein
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : x,y             - Position des Pixel des neuen Elements
//                      wert            - Farbwert des Pixels des neuen Elements
//                      si              - si-Wert der Kachel des neuen Pixels
//  Ausgangsparameter : first,last      - Zeiger auf Anfang und Ende der Liste
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : -
// =================================================================================================
procedure AddElement(x,y : longint; wert,si : byte; var first,last : TStartpunktPtr);
var
 p : TStartpunktPtr;
begin
 new(p);                               // erstelle neues Element und flle es mit den Daten
 p^.x       := x;
 p^.y       := y;
 p^.wert    := wert;
 p^.si      := si;
 p^.prev    := last;
 p^.next    := nil;

 if last <> nil then                   // verbiege Lastzeiger, wenn Liste nicht leer war
   last^.next := p;

 if first = nil then                   // verbiege First-Zeiger, wenn Liste zuvor leer
   first := p;

 last := p;                            // Last-Zeiger zeigt auf eingefgtes, neues letztes Element
end;


// =================================================================================================
//  Funktion          : Entfernt alle Elemente aus den Listen und gibt den Speicher wieder frei
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : -
//  Ausgangsparameter : first,last      - Zeiger auf Anfang und Ende der globalen Liste
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_first, g_last
// =================================================================================================
procedure LoescheListe(var first : TStartpunktPtr; var last : TStartpunktPtr);
var
 i   : longint;
 del : TStartpunktPtr;
begin
 for i := 0 to Length(g_Bereiche)-1 do      // durchlaufe Bereichsarray
 begin
   while g_bereiche[i].First <> nil do    // durchlaufe die gesamte Liste eines Bereiches
   begin
      del := g_bereiche[i].First;         // entferne aktuelles Element
      g_bereiche[i].First := g_bereiche[i].First^.next;
      Dispose(del);
   end;
 end;
 first := nil;                            // First und Last-Zeiger wieder rcksetzen
 last  := nil;
end;


// =================================================================================================
//  Funktion          : prft, ob ein Startpunkt gelscht werden kann, wenn auf der horizontalen
//                      bei jedem Pixel die Intensittsschwelle nicht berschritten wird
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : x,y                     - Position des Pixel des neuen Elements
//                      ziel                    - x-koord des Zielwertes
//                      grenzeOben, grenzeUnten - Intensittsintervall, innerhalb dessen der
//                                                jeweilige Pixelwert liegen muss
//                      
//  Ausgangsparameter : Result                  - Weg vorhanden oder nicht
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild
// =================================================================================================
function WegHorizontal(x,y,ziel,i,grenzeOben,grenzeUnten : Integer) : Boolean;
begin
 Result := False;
 while x <> ziel do
 begin
   if (g_bild[x,y].wert < grenzeUnten) or (g_bild[x,y].wert > grenzeOben) then
      exit;                            //verlasse fkt, wenn Schwelle zu gro ist
   x := x + i;
 end;
 Result := True;
end;


// =================================================================================================
//  Funktion          : prft, ob ein Startpunkt gelscht werden kann, wenn auf der Vertikalen
//                      bei jedem Pixel die Intensittsschwelle nicht berschritten wird
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : x,y                     - Position des Pixel des neuen Elements
//                      ziel                    - y-koord des Zielwertes
//                      grenzeOben, grenzeUnten - Intensittsintervall, innerhalb der jeweilige 
//                                                Pixelwert liegen muss
//                      
//  Ausgangsparameter : Result                  - Weg vorhanden oder nicht
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild
// =================================================================================================
function WegVertikal(x,y,ziel,i,grenzeOben,grenzeUnten : Integer) : Boolean;
begin
 Result := False;
 while y <> ziel do
 begin
   if (g_bild[x,y].wert < grenzeUnten) or (g_bild[x,y].wert > grenzeOben) then
      exit;                            //verlasse fkt, wenn Schwelle zu gro ist
   y := y + i;
 end;
 Result := True;
end;


// =================================================================================================
//  Funktion          : prft on ein Weg zwischen zwei Punkten existiert, auf dem die Intesitts-
//                      schwelle nicht berschritten wird
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : p1,p2                   - die beiden Punkte, die geprft werden sollen
//  Ausgangsparameter : Result                  - Weg vorhanden oder nicht
// =================================================================================================
function FindeWeg(p1, p2 : TStartpunktPtr) : boolean;
var
 si2           : Integer;
 grenzeOben,
 grenzeUnten   : Integer;
 i1,i2         : Integer;
const
 C_I           : array[0..1] of Integer = (-1, 1);
begin
 si2           := Trunc(p1^.si/2);           //definiere obere und untere Intensittsgrenze
 Result        := False;
 grenzeOben    := p1^.wert + si2;
 grenzeUnten   := p1^.wert - si2;

 // Ist Zielpixel im gleichen Farbbereich
 if (p2^.wert < grenzeUnten) or (p2^.wert > grenzeOben) then
   exit;

 // erst horizontal, dann vertikal und umgekehrt
 i1 := C_I[ Ord(p1^.x < p2^.x) ];
 i2 := C_I[ Ord(p1^.y < p2^.y) ];
 if ( WegHorizontal(p1^.x, p1^.y, p2^.x, i1, grenzeOben, grenzeUnten) and WegVertikal(p2^.x, p1^.y, p2^.y, i2, grenzeOben, grenzeUnten)   ) or
    ( WegVertikal(p1^.x, p1^.y, p2^.y, i2, grenzeOben, grenzeUnten)   and WegHorizontal(p1^.x, p2^.y, p2^.x, i1, grenzeOben, grenzeUnten) ) then
 begin
   Result := True;
   exit;
 end;

end;


// =================================================================================================
//  Funktion          : rastert die mglichen Startpunkte so, dass in x und y Richtung nur jeder
//                      "teiler-te" Punkt als Startpunkt in Frage kommt - Geschwindigkeitsoptimierung
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : teiler   - bestimmt die Rastergre
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild
// =================================================================================================
procedure Aussieben(teiler : longint);
var
 x,y     : Longint;
begin
 for y:=1 to g_hoehe-2 do           // Da die Bildrandpunkte eh schon alle markiert sind, werden diese nicht 
   for x:=1 to g_breite-2 do        // durchlaufen
      if (x mod teiler <> 0) or (y mod teiler <> 0) then
         g_bild[x, y].todel := true;
end;


// =================================================================================================
//  Funktion          : prft jeden (nach der Rasterung) mglichen Startpuunkt mit jedem, ob sie
//                      ausgelscht werden knnen. Die briggebliebenen Punkte bilden die Saatpunkte
//                      fr das eigentliche Bereichswachstum
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_first, g_last, 
//                      g_ticks, g_strStatus
// =================================================================================================
procedure LoeschePunkte;
var
 p1, p2, del : TStartpunktPtr;
begin
 if g_first = nil then
   exit;

 // Zeitmessung
 g_ticks.markieren := timeGetTime;

 p1 := g_first;
 p2 := g_first^.next;
 while (p1^.next <> nil) do
 begin
   p2 := p1^.next;                 // diese Schleifenschachtelung ermglicht den Vergleich aller Punkt-
   while p2 <> nil do              // kombinationen in einer linearen Liste
   begin                            
      del := p2;
      p2 := p2^.next;
      if FindeWeg(p1, del) then
         LoescheElement(del, g_first, g_last);  //lsche den Punkt aus der Liste, der vom anderen erreicht worden ist
   end;
   if p1^.next <> nil then
      p1 := p1^.next;
 end;
 g_ticks.markieren := timeGetTime - g_ticks.markieren;
 g_strStatus  := g_strStatus+ #13#10+'Zeit fr das Markieren der Startpunkte (in ms):'+#09+#09+IntToStr(g_ticks.markieren);
end;


// =================================================================================================
//  Funktion          : prft fr das Bereichswachstum, ob ein Punkt in einen Bereich aufgenommen wird
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : x,y         - Position des Pixel des zu prfenden Punktes
//                      min, max    - Intensittsintervall, innerhalb dessen der jeweilige 
//                                    Pixelwert liegen muss
//                      aktBereich  - der Bereich, in den der Pixel evtl. aufgenommen wird
//                      k           - Zhler, der angibt, wie viele nicht zugeordnete Punkte noch
//                                    im Bild vorhanden sind
//  Ausgangsparameter : Result      - Pixel entweder vorher schon in aktBereich oder jetzt neu dazu
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_breite, g_hoehe
// =================================================================================================
function pruefeNachbar (min, max : byte; x,y : longint; var aktBereich : TBereich; var k : Longint) : boolean;
begin
   if (x >= 0) and (y >= 0) and (x < g_breite) and (y < g_hoehe)                            //pixel in bild?
                 and (g_bild[x,y].Bereich = C_LEERER_BEREICH)                               //nachbar nicht zugeprdnet?
                 and (g_bild[x,y].wert >= min) and (g_bild[x,y].wert <= max) then           //Intensittsdifferenz ok?
   begin
      AddElement(x,y, g_bild[x,y].wert, g_bild[x,y].si, aktBereich.First, aktBereich.Last); //fge Nachbarpunkt in Startpunktliste ein
      inc(aktBereich.FarbwertSumme, g_bild[x,y].wert);                                       
      inc(aktBereich.AnzPunkte);
      g_bild[x,y].Bereich := aktBereich.id;
      dec(k);
      aktBereich.Wachsbar := true;           // wenn Punkt hinzugekommen ist, kann der Bereich theoretisch noch weiter wachsen
   end;
 if (x >= 0) and (y >= 0) and (x < g_breite) and (y < g_hoehe) then                         //pixel in bild?
   Result := g_bild[x,y].Bereich = aktBereich.id
 else
   Result := True;

end;


// =================================================================================================
//  Funktion          : fhrt das eigentliche Bereichswachstumsverfahren durch
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : deltaI            - Abweichung, die ein Pixel beim ersten Durchlauf hchstens
//                                          haben darf, um in Bereich aufgenommen zu werden
//                      toleranzerhoehung - Wert, um den die toleranz bei jedem Durchlauf erhht wird
//                      fastmode          - Modus, ob Bereiche seriell (fast) oder pseudoparallel
//                                          (slow) wachsen 
//  Ausgangsparameter : -
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_breite, g_hoehe, g_bereiche, g_first, 
//                      g_ticks, g_strStatus
// =================================================================================================
procedure Wachsen(deltaI, toleranzerhoehung : integer; fastmode : boolean);
var
 p, del, last_alt    : TStartpunktPtr;
 last_alt_flag       : boolean;
 AnzBereiche         : longint;
 i, k                : longint;
 min,max             : integer;
 ausgewachsen        : boolean;
 ende_flag           : boolean;
begin
 // Zeitmessung
 g_ticks.wachsen := timeGetTime;

 // Setze Bildpixel aller Startpunkte auf unterschiedl. Bereiche (kein Bereich: -1)
 i:=0;
 p := g_first;
 while (p <> nil) do
 begin
   SetLength(g_bereiche, i+1);                  //incrementiere die Arraylnge bei jedem Schleifendurchlauf
   g_bereiche[i].FarbwertSumme   := p^.wert;    
   g_bereiche[i].AnzPunkte       := 1;
   g_bereiche[i].Wachsbar        := true;
   g_bereiche[i].First           := p;
   g_bereiche[i].Last            := p;          //initialisiere Bereich
   g_bereiche[i].id              := i;
   g_bild[p^.x, p^.y].Bereich    := i;
   p                             := p^.next;
   g_bereiche[i].First^.prev     := nil;
   g_bereiche[i].First^.next     := nil;
   inc(i);
 end;

 // Zhler fr nicht zugeordnete Bildpunkte
 k := g_breite*g_hoehe - i;
 AnzBereiche := i;

 //ShowMessage('Bereiche: '+IntToStr(i) );

 repeat                                //laufe so oft durch, bis alle Punkte des Bildes zugeordent sind

   ausgewachsen := false;
   while not ausgewachsen do           //while mind. ein Bereich wachsen kann
   begin
      i := 0;
      for i := 0 to AnzBereiche-1 do   //durchlaufe Bereiche
      begin
         //ShowMessage('for1!');
         g_bereiche[i].Wachsbar := false;
         p := g_bereiche[i].First;
         last_alt := g_bereiche[i].last;
         last_alt_flag := false;
         repeat                        //durchlaufe bis Bereiche zuendegewachsen ist (fastmode), oder
                                       //Bereichswechsel wenn pseudoparallelitt stattfindet
            //if p^.next=nil then
              //ShowMessage('repeatanfang: p^.next = nil');

            min := Round(g_bereiche[i].FarbwertSumme / g_bereiche[i].AnzPunkte - deltaI);
            max := Round(g_bereiche[i].FarbwertSumme / g_bereiche[i].AnzPunkte + deltaI);
            if min < 0 then
               min := 0;
            if max > 255 then
               max := 255;
            {$B+}
            if pruefeNachbar(min, max, p^.x-1, p^.y-1, g_Bereiche[i], k) and
               pruefeNachbar(min, max, p^.x,   p^.y-1, g_Bereiche[i], k) and
               pruefeNachbar(min, max, p^.x+1, p^.y-1, g_Bereiche[i], k) and
               pruefeNachbar(min, max, p^.x-1, p^.y,   g_Bereiche[i], k) and
               pruefeNachbar(min, max, p^.x+1, p^.y,   g_Bereiche[i], k) and
               pruefeNachbar(min, max, p^.x-1, p^.y+1, g_Bereiche[i], k) and
               pruefeNachbar(min, max, p^.x,   p^.y+1, g_Bereiche[i], k) and
               pruefeNachbar(min, max, p^.x+1, p^.y+1, g_Bereiche[i], k) then
            begin
               {$B-}                   //wenn alle Nachbarpunkte eines Punktes im Bereich drin sind ist der Punkt
               del := p;               //kein Bereichsrandpunkt und kann aus der Liste gelscht werden
               //if p=nil then
                 //ShowMessage('then: p = nil');

               p := p^.next;
               if del = last_alt then
                  last_alt_flag := true;
               if AnzBereiche > 1 then
                  LoescheElement(del, g_bereiche[i].First, g_bereiche[i].Last);    //entferne p aus liste
            end
            else
            begin
               {$B-}
               //if p=nil then
                 //ShowMessage('else: p = nil');

               p := p^.next;
            end;


            if fastmode then
               ende_flag := p=nil            //fastmode: Bereiche wachsen seriell
            else
               ende_flag := last_alt_flag or (p = last_alt^.next);
                                             //slowmode: Bereiche wachsen pseudoparallel
             //if ende_flag then
               //ShowMessage('repeat2 ende!');
         until ende_flag;
      end;
     ausgewachsen := true;
      for i := 0 to AnzBereiche-1 do         //prfe, ob alle Bereiche ausgewachsen sind
         ausgewachsen := ausgewachsen and not g_bereiche[i].Wachsbar;
   end;
 inc(deltaI,toleranzerhoehung);              //erhhe Toleranz fr nchsten durchlauf
 //ShowMessage('k: '+IntToStr(k));
 until (k=0) or (AnzBereiche=0);                                  //wenn k = 0: Alle Punkte sind zugeordnet

 g_ticks.wachsen   := timeGetTime - g_ticks.wachsen;
 g_strStatus       := g_strStatus+ #13#10+'Zeit fr das Wachsen der Bereiche (in ms):'+#09+#09+IntToStr(g_ticks.wachsen);
 g_strStatus       := g_strStatus+ #13#10+'Anzahl Bereiche vorm Verschmelzen:'+#09+#09+#09+IntToStr(Length(g_bereiche) );
end;


// =================================================================================================
//  Funktion          : prft, ob zwei Bereiche im Bild nebeneinander sind bzw. ob sie aneinander
//                      angrenzen
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : i,j      - ID (Index) der beiden Bereiche
//  Ausgangsparameter : Result   - gibt zurck, ob die Bereiche benachbart sind
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bereiche, g_breite, g_hoehe
// =================================================================================================
function BereicheNebeneinander(i,j : longint) : Boolean;
var
 p                : TStartpunktPtr;
 x,y              : longint;
begin
 Result := False;
 p := g_bereiche[i].First;
 while (p <> nil) do
 begin
   // durchlaufe alle Nachbarpunkte jedes Punktes des ersten Bereiches
   for x:=p^.x-1 to p^.x+1 do
      for y:=p^.y-1 to p^.y+1 do
         // wenn Nachbarpixel im Bild und Nachbarpixel im zweiten Bereich ist -> Bereiche benachbart!
         if (x>=0) and (x<g_breite) and (y>=0) and (y<g_hoehe) and (g_bild[x,y].Bereich=g_bereiche[j].id) then
         begin
            Result := True;
            exit;
         end;

   p := p^.next;
 end;
end;


// =================================================================================================
//  Funktion          : ermittelt die Anzahl gltiger Bereich in dem Bild und gibt sie zurck
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : -
//  Ausgangsparameter : Result   - gibt die Anzahl von Bereichen zurck
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bereiche
// =================================================================================================
function AnzahlBereiche : Longint;
var
 i : integer;
begin
 Result := 0;
 for i:=0 to Length(g_bereiche)-1 do            // durchlaufe Bereichsarray
   if (g_bereiche[i].id > -1) then              // wenn Bereich gltig -> inkrem. Zhler
      inc(Result);
end;

// =================================================================================================
//  Funktion          : durchluft und berprft benachbarte Bereiche auf hnliche Farbintensitten
//                      und verschmilzt hnliche Bereiche miteinander zu neuen greren Bereichen
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : toleranz                - Toleranzwert fr den Intens.-Unterschied zweier
//                                                Bereiche, die noch zusammengefasst werden
//
//  Ausgangsparameter : -
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_ticks, g_strStatus, g_bereiche
// =================================================================================================
procedure Verschmelzen(toleranz : byte);
var
 p                : TStartpunktPtr;
 i,j,x,y          : longint;
begin
 // Zeitmessung
 g_ticks.verschmelzen := timeGetTime;

 // durchlaufe alle Bereiche und vergleiche immer zwei miteinander
 for i:=0 to Length(g_bereiche)-2 do
   for j:=i+1 to Length(g_bereiche)-1 do
      // wenn beide Bereiche gltig und deren Intensitten in der Toleranz und beide Bereiche benachbart sind
      if (g_bereiche[i].id > -1) and (g_bereiche[j].id > -1) and
         (Abs( Round(g_bereiche[i].FarbwertSumme / g_bereiche[i].AnzPunkte)-
              Round(g_bereiche[j].FarbwertSumme / g_bereiche[j].AnzPunkte) ) <= toleranz) and
         BereicheNebeneinander(i,j)  then
      begin
            // verschmelze die beiden Bereiche
            g_bereiche[j].FarbwertSumme   := g_bereiche[j].FarbwertSumme + g_bereiche[i].FarbwertSumme;
            g_bereiche[j].AnzPunkte       := g_bereiche[j].AnzPunkte + g_bereiche[i].AnzPunkte;
            g_bereiche[j].Last^.next      := g_bereiche[i].First;
            g_bereiche[i].First^.prev     := g_bereiche[j].Last;
            g_bereiche[j].Last            := g_bereiche[i].Last;

            // gehe alle Bildpixel durch und passe sie an neuen zusammengefassten Bereich an
            // (in der g_bereiche[]-Liste sind nur die Randpunkte der Bereiche enthalten )
            for y:=0 to g_hoehe-1 do
               for x:=0 to g_breite-1 do
                  if (g_bild[x,y].Bereich = g_bereiche[i].id) then
                     g_bild[x,y].Bereich := g_bereiche[j].id;

            // zweiter Bereich wird berflssig und wird als 'ungltig' markiert
            g_bereiche[i].id              := -1;
            g_bereiche[i].First           := nil;
            g_bereiche[i].Last            := nil;
            g_bereiche[i].FarbwertSumme   := 0;
            g_bereiche[i].AnzPunkte       := 0;
            g_bereiche[i].Wachsbar        := false;
      end;

 // berechne die neue Anzahl an Bereichen und die verstrichene Rechenzeit des Verschmelzvorgangs
 g_ticks.verschmelzen := timeGetTime - g_ticks.verschmelzen;
 g_strStatus          := g_strStatus+ #13#10+'Anzahl Bereiche nach dem Verschmelzen:'+#09+#09+IntToStr( AnzahlBereiche );
 g_strStatus          := g_strStatus+ #13#10+'Zeit fr das Verschmelzen der Bereiche (in ms):'+#09+#09+IntToStr(g_ticks.verschmelzen);
end;


// =================================================================================================
//  Funktion          : Initialisiert alle Filter-Parameter mit Standardwerten, wenn keine dem
//                      Plugin bergeben wurden
// -------------------------------------------------------------------------------------------------
//  Ausgangsparameter : FilterKern              - enthlt die bergabeparameter an das Plugin
//                      RGBVektor1              - Optionale Parameter vom Bediener
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_ticks, g_strStatus, g_hoehe, g_breite, g_bereiche
// =================================================================================================
procedure InitFilterKern(var FilterKern : TFilterKern; var RGBVektor1 : TRGBVektor);
var
 i,j : integer;
begin
 for i:=0 to 2 do
   for j:=0 to 2 do
   begin
      if Round(FilterKern[i,j]) = 0 then
         FilterKern[i,j] := C_DEFAULT[i,j];
      if Round(FilterKern[i,j]) < C_MIN[i,j] then
         FilterKern[i,j] := C_MIN[i,j];
      if Round(FilterKern[i,j]) > C_MAX[i,j] then
         FilterKern[i,j] := C_MAX[i,j];
   end;
end;


// =================================================================================================
//  Funktion          : Enthlt die Hauptroutine des Filters und wird in dieser DLL exportiert.
//                      Die Funktion prft und setzt alle an den Filter bergebenen Parameter
//                      und wandelt das bergebene Bild (falls ntig) von RGB-Farben in ein
//                      Graustufenbild um. Danach werden Initialpunkte gesucht und das
//                      Wachstumsverfahren mit anschlieender Verschmelzung der gefunden Bereiche
//                      durchgefhrt.
// -------------------------------------------------------------------------------------------------
//  Eingangsparameter : pSrc1                   - Zeiger auf erstes Quellbild (auf das der Filter
//                                                angewendet werden soll)
//                      pSrc2                   - Zeiger auf zweites Quellbild (hier unbenutzt)
//                      dst                     - Referenz auf das Zielbild
//                      fk                      - enthlt die bergabeparameter an das Plugin
//                      RGBVektor1/2            - Optionale Parameter vom Bediener
//
//  Ausgangsparameter : Result                  - Statuscode
// -------------------------------------------------------------------------------------------------
//  Globale Zugriffe  : g_bild, g_ticks, g_strStatus, g_hoehe, g_breite, g_bereiche
// =================================================================================================
function ImageProcess(pSrc1, pSrc2:Pointer; dst:TCVImage; fk:TFilterKern; RGBVektor1, RGBVektor2:TRGBVektor) : Longint; stdcall;
var
 lDstBase         : longint;
 pDstVPA          : pVPAT;
 x,y              : longint;
begin
 // wenn kein Bild angegeben wurde -> Abbruch
 if not Assigned(pSrc1) then
 begin
   g_strStatus  := 'ImageProcess() - Bitte ein primres Bild auswhlen!';
   ImageProcess := -1;
   exit;
 end;
 // Zeitmessung
 g_ticks.summe := timeGetTime;

 // wenn Graubild bergeben->einfach in Ziel kopieren, sonst von RGB in Grau umwandeln
 if (ImageDimension (pSrc1) = 1) then
   KopiereGraubild(pSrc1)
 else
   KopiereFarbbild(pSrc1);


 InitFilterKern(fk, RGBVektor1);


 BerechneIntSchwellen(Round(fk[0,0]),Round(fk[0,1]),C_INTK);            // Intensittsschwellen der Quadrate berechnen
 BerechneGradient(Round(fk[1,0]), TGradient( Round(fk[1,1]) ) );        // Randpunkte finden und aussortieren
 Aussieben(Round(fk[0,2]));                                             // Geschw.-Optimierung: Nur jeden x-ten Pixel als Startpkt verwenden
 ErzeugeListe;                                                          // Liste aller brigen, mglichen Startpunkte erzeugen
 LoeschePunkte;                                                         // entfernt alle berflssigen Startpunkte, die auf geradem
                                                                        // Weg (horiz./vertikal zu den Achsen) von anderen Pkt. erreichbar sind
 Wachsen(Round(fk[1,2]), Round(fk[2,0]), Round(fk[2,1])=1);             // Fhrt das Region Growing an den gefunden Startpunkten durch
 Verschmelzen(Round(fk[2,2]));                                          // fasst nebeneinanderliegende Bereiche zusammen, wenn Ihre
                                                                        // Intensitten sehr hnlich sind (verschmilzt benachbarte Bereiche)

 // durchlaufe alle Pixel und setze Farbe aus den Bereichsdaten des Pixels
 for y:=0 to g_hoehe-1 do
   for x:=0 to g_breite-1 do
      if g_bild[x,y].Bereich <> C_LEERER_BEREICH then     // ungltiger Bereich (fr DEBUGZWECKE, kommt nicht vor)
         g_bild[x,y].wert := Round(g_bereiche[ g_bild[x,y].Bereich ].FarbwertSumme /
                                   g_bereiche[ g_bild[x,y].Bereich ].AnzPunkte);

 // kopiert das fertige Ergebnisbild zurck in die Bild-Struktur des Plugins
 KopiereZielbild(dst);
 LoescheListe(g_first, g_last);

 // berechne anhand vergangener Ticks die bentigte Rechenzeit dieses Filters
 g_ticks.summe := timeGetTime - g_ticks.summe;
 g_strStatus  := g_strStatus+ #13#10+'Zeit fr kompletten Filter (in ms):'+#09+#09+#09+IntToStr(g_ticks.summe);

 ImageProcess := 0;
end;


// ============================================================================
//  Name      				: GetInfo
// ----------------------------------------------------------------------------
//  Eingangsparameter : -					 | -
//                                      |
//  Ausgangsparameter : RESULT			 | Infostruktur mit konstanten Strings,
//													 | Typ: TInfoString
// ----------------------------------------------------------------------------
//  Funktion          : Die vier Stringkonstanten werden im Memofeld des
//											MDI-Ausgabefensters (MDI-Zieldokument) angezeigt.
//  Bemerkungen       : -
// ============================================================================
function GetInfo: TInfoString; stdcall;
begin
 GetInfo.pFilterGroup   := 'Gruppe 6';
 GetInfo.pFilterName    := 'Region Growing';
 GetInfo.pFilterAuthor  := 'Sebastian Reuss (ii5062) und Andreas Schibilla (ii4900)';
 GetInfo.pFilterVersion := '0.23';
end;


// ============================================================================
//  Name    				: GetStatusString
// ----------------------------------------------------------------------------
//  Eingangsparameter : -				    | -
//                                      |
//  Ausgangsparameter : RESULT			 | Pointer auf Huge-String (PChar).
// ----------------------------------------------------------------------------
//  Funktion          : Stellt zustzliche Statusinformationen zur Verfgung,
//											die im Memofeld des MDI-Ausgabefensters
//											(Ergebnisbild) angezeigt werden.
//  Bemerkungen       : Wird erst ab BVATisch-Version 2.0.0.0 ausgewertet.
// ============================================================================
function GetStatusString: PChar; stdcall;
begin
	Result := PChar(g_strStatus);
end;


// ============================================================================
// Die folgenden Routinen werden von der DLL exportiert und mssen auch
// von den neuen (im Praktikum programmierten) Plugins zur Verfgung gestellt
// werden:
// - "ImageProcess" fhrt die eigentliche Filteroperation aus.
// - "GetInfo" liefert Versionsinfos zum Plugin und wird im MemoFeld des
//		zugeordneten MDI-Childs dargestellt.
// - "GetStatusString" liefert zustzliche Infos aus "ImageProcess" an die
//		Hostanwendung BVATisch (>= V2.0.0.0 ist ntig)
// ============================================================================
exports
 ImageProcess, GetInfo, GetStatusString;


begin
end.
